/*
  
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/
  
  ===================================================
  
  Name:       CCM (bookKeeping.C)
  
  Copyright:  Damian J. J. Farnell
  
  Date:       Wed Sep 29 1999
  
  Purpose:    Determine and solve CCM LSUBm equations.

  Rewritten:  Joerg Schulenburg 2005
              - cout replaced by printf (was mixed with printf)
  
  ===================================================
  
  ---------------------------------------------------

  MODULE NAME: BOOKKEEPING.C 

  PURPOSE:     This module does all of the `bookkeeping' 
               for all of the other modules. Hence, all
               routines that are multiply used throughout
               the code are contained here. 

  ---------------------------------------------------

*/


// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#include "../include/nr.h"
#include "../include/nrutil.h"
#include "../include/Constants.h"
#include "../include/Externals.h"
#include "../include/FunctionPrototypes.h"

int findGeneralisedPositionFromCartesians(
         double x, double y, double z, int *configuration);

// return the number of realtime seconds since start
int seconds( void ) {
  static time_t start_time=0;
  time_t t1;
  time(&t1); if (start_time==0) start_time=t1;
  return (t1-start_time);
}

void printOutConfig(int (*conf)[5], int order) {
  int i,j,max[3];
  static int animal[MAX_LSUBM+1][MAX_LSUBM+1];
  max[0]=max[1]=max[2]=0; /* covering cube */

  for (i=0;i<MAX_LSUBM+1;i++)
    for (j=0;j<MAX_LSUBM+1;j++)
      animal[i][j]=0;
  printf("(x,y,z,unitcell):");
  for (i=1; i<=order; i++) {
    printf(" (%d,%d,%d,%d)", conf[i][1], conf[i][2], conf[i][3], conf[i][4]);
    if (max[0]<conf[i][1]) max[0]=conf[i][1];
    if (max[1]<conf[i][2]) max[1]=conf[i][2];
    if (max[2]<conf[i][3]) max[2]=conf[i][3];
    animal[ conf[i][2] ][ conf[i][1] ]++;
  }
  printf("\n");
  if (max[2]<1) { /* flat animal? */
      for (i=0;i<max[1]+1;i++) {
        printf(" ");
        for (j=0;j<max[0]+1;j++)
          printf("%c",((animal[i][j])?'0'+animal[i][j]:'.'));
        printf("\n");
      }
  }
}

void printOutDescriptor(int *descriptor) {
  int i;
  printf("descriptor[1..%d]:  ",MAX_SPIN_LIST-1);
  for(i=1;i<MAX_SPIN_LIST;i++) 
    printf("%d ",descriptor[i]);
  printf("\n");
}

/* convert old obsolete array-format to new format and call printOutConfig */
void printOutCluster(int length, int **config) {	
  int l;
  static int conf[MAX_LSUBM+1][5];
  for(l=1;l<=length;l++) {   // x  y   z   UC.number
    conf[l][1]=config[l][1];
    conf[l][2]=config[l][2];
    conf[l][3]=config[l][3];
    conf[l][4]=config[l][4];
  }
  printOutConfig(conf,length);
}


// storage place for (Params.s2+1)^(j-1) to speedup code
// sum of all divisors must fit into integer (for spin-bit pattern)
int *divisors=NULL;  // multiplicators (2s+1)^(i-1) for each site i
int block_size=0;    // number of sites fitting to a integer

void initDivisors(void) {  // Oct05: rewritten
  int tempS, j, prod;
  tempS = Params.s2+1;  // 2s+1
  if (tempS<2) { printf("Error initDivisors: 2s<1\n"); exit(1); }

  // calculate the maximum number of sites fitting to an integer
  for(j=1,prod=1; ; prod*=tempS, j++) {   // tempS ^ (j-1)
    if ((prod * tempS) * 1.0 != (prod * 1.0) * tempS) break;
  }
  if (myid==0)
    printf("initDivisors: maximum sites fitting to int-size is %d\n",j);
  block_size=j-1;    // BLOCK_LENGTH = j-1;

  // MAX_SPIN_LIST*block_size >= MAX_BT*uc // ToDo: check max_spin_list
  divisors = (int *) malloc(sizeof(int)*(block_size+1));
  if (!divisors) { printf("Error initDivisors: malloc failed\n"); exit(1); }
  divisors[1] = 1;
  for(j=1, prod=1; j<=block_size; prod*=tempS, j++) {
    divisors[j] = prod;  // = (2s+1) ^ (j-1)
  }
}


// in: order of ccm (maximum spinflips)
//     configuration[1...order?][5]
// out: descriptor[MAX_SPIN_LIST] of smallest symmetric config
// cost most CPU time (Joerg Oct2004)
// called by findFundamentalSpinConfigurations() in define.C
//       and termXX() in obtain.C
// return 1, if symmetry operations can not lower the descriptor (Nov05)
int findUniqueDescriptorOfCluster(int order,
				   int **configuration,
				   int *descriptor){

  /* This routine gets a set of crystallographic lattice sites
     which define a cluster and returns a set of unique integer 
     numbers in `descriptor' which fully describe it.

     Note that this is its FUNDAMENTAL description, i.e., 
     of the how ever many possible symmetry equivalent 
     cluster there are of it, this `descriptor' has 
     the lowest numerical value.

     PARTS ARE:

     1) determine symmetry related equivalents and translate
     ...into the x>=0,y>=0,z>=0 part.

     2) determine the `spin' form of these equivalents

     3) turn this `spin' form into a unique set of integers

     4) find set of integers with lowest `numerical' value
     ...out of all the symmetry related clusters and return
     ...this in `descriptor'.

  */

  int i,j,k;
  int minR[3];  // x y z
  int OK, is_minConfig;
  static int n_out=0;  // ToDo: reset if &(*configuration) changes
  // int **tempConfiguration;  // obsolete
  static int tempDescriptor[MAX_SPIN_LIST];

  // tempConfiguration=matrix(0,MAX_LSUBM+1,0,5);  // to slow
  static int tempConfiguration[MAX_LSUBM+1][5];

  is_minConfig=1;
  for(k=1;k<MAX_SPIN_LIST;k++)
    descriptor[k]=999999; // Oct05  was 10000000
#if 1
  for(j=1;j<=order;j++) for (k=1; k<5; k++) {
      tempConfiguration[j][k]= configuration[j][k];
      if (configuration[j][k]<0) is_minConfig=0;
  }
  if (is_minConfig)  // will fail for negativ koordinates
    findConfigurationToDescriptor(descriptor, tempConfiguration, order);
  // printOutDescriptor( descriptor );
#endif
  
  if (n_out<=100) n_out++;

  if (Control.output >4 && n_out<100)
      printf("findUniqueDescriptorOfCluster:\n");
  
  for(i=1;i<=LatticeSymmetries.number;i++) {

    if (Control.output >5 && n_out<100) {
      printf("Symmetry %d of %d:\n",i,LatticeSymmetries.number);
    }
    
    // find symmetry related clusters

    symmetryGroup(i,order,configuration,tempConfiguration);

    if (Control.output >6 && n_out<100) {
      printf("transformed to:\n");
      printOutConfig(tempConfiguration, order);
    }
    
    // translate into correct quadrant, x>=0,y>=0,z>=0
    
    for (k=0; k<3; k++) minR[k]=10000000;

    for(j=1;j<=order;j++)
      for (k=0; k<3; k++)
        if (tempConfiguration[j][k+1]<minR[k]) 
           minR[k] = tempConfiguration[j][k+1];

    for(j=1;j<=order;j++)
      for (k=0; k<3; k++)
        tempConfiguration[j][k+1] -= minR[k];

    if (Control.output >6 && n_out<100){
      printf("translated to:\n");
      printOutConfig(tempConfiguration, order);
    }

    // now find the descriptor (costs most CPU, Joerg Oct04)
    // bitpattern (flips per site.uc) splittet to an integer array
    
    findConfigurationToDescriptor(tempDescriptor, tempConfiguration, order);
    // if tempDescriptor[]=999999 we have a invalid cluster

    if (Control.output >5 && n_out<100) {
      printf("Temp. Descriptor(%d) = ",i); // i = number of lattice symmetry
      printOutDescriptor( tempDescriptor );
    }

    if (tempDescriptor[1]==999999) { // Joerg Oct05
       if (Control.output>7) // term 27 + term 30
         printf("Error findUniqueDescriptorOfCluster: invalid descriptor,"
              " sym=%d\n", i);
       continue;
    }

    // check if tempDesc is smaller than Desc

    for (OK=0,k=(MAX_SPIN_LIST-1); k>0; k--) {
      if (tempDescriptor[k] > descriptor[k]) { OK=-1; break; }
      if (tempDescriptor[k] < descriptor[k]) { OK= 1; break; }
    }

    if (OK==1) {
      if (descriptor[MAX_SPIN_LIST-1]!=999999)
        is_minConfig = 0; /* input config will be replaced */
      for(k=1;k<MAX_SPIN_LIST;k++)
	descriptor[k]=tempDescriptor[k];
    }

  }

#if DEBUG
  if (descriptor[1]==999999) {
    printf("Can't find a reasonable descriptor of this cluster!\n"
      "Exiting...\n\n");
    exit(1);
  }
#endif

  if (Control.output > 4 && n_out<100) {
    printf(" Minimum Descriptors = ");
    printOutDescriptor( descriptor );
  }
  // free_matrix(tempConfiguration,0,MAX_LSUBM+1,0,5);

  return is_minConfig; /* Joerg Nov05: for counting configs with low memory */
 
}

// in:  sym=symmetry_number
//      order - length of configuration (ToDo: use stop code? config[4]<0?)
//      configuration[1..order][1..4] 
// out: tempConfiguration[1..order][1..4] - symmetric coords 
void symmetryGroup(int sym, 
		   int order,
		   int **configuration,
		   int (*tempConfiguration)[5])
{
  int i;

  double newx, newy, newz;
  double x,y,z;

  // symmorphic point group symetries entered here only =>
  // triangular lattice not-yet implemented, to be done.

  if (sym > 48 || sym < 1) {
    printf("Switch Variable out of range %d...\nExiting!\n\n", sym);
    exit(1);
  }

  for(i=1;i<=order;i++) {
    
    // configuration[5] -> (x,y,z)
    findCartesiansFromGeneralisedPosition(x, y, z, configuration[i]);
    
    newx=x*LatticeSymmetries.R[sym][1][1]+
         y*LatticeSymmetries.R[sym][1][2]+
         z*LatticeSymmetries.R[sym][1][3]+
           LatticeSymmetries.x[sym];
    
    newy=x*LatticeSymmetries.R[sym][2][1]+
         y*LatticeSymmetries.R[sym][2][2]+
         z*LatticeSymmetries.R[sym][2][3]+
           LatticeSymmetries.y[sym];
    
    newz=x*LatticeSymmetries.R[sym][3][1]+
         y*LatticeSymmetries.R[sym][3][2]+
         z*LatticeSymmetries.R[sym][3][3]+
           LatticeSymmetries.z[sym];
  
    // ToDo: this consumes most of the time Sep05
    // restore tempConfig=(x,y,z,uc) from (newx,newy,newz) ToDo: use new[3]
    if ( findGeneralisedPositionFromCartesians(
                  newx,newy,newz,tempConfiguration[i])) {
       // Fatal Error: not found
       printf("Error symmetryGroup: sym=%d site=%d\n", sym, i);
       printOutCluster(order, configuration);
       // exit(1); // Oct05 ToDo: probably its just because of inaccuracy
    }
  }
}

// in:  x y z
// out: configuration[1..4]
// called: by symmetryGroup()  
// return 0 on success, otherwise fatal error
// ToDo: called very often by obtain! takes much cpu time, Joerg Sep05
int findGeneralisedPositionFromCartesians(double x, double y, double z, 
					   int *configuration)
{
  static int index[5];
  int i, j, k;

  // double **m, **tempm;
  // double *s;
  double distance, mindist=-1;
  // int *ind;

  double d;

  static int  *ind=ivector(0,5);    // call it only once for faster code (Joerg)
  static double *s=dvector(0,5);
  static double **m=dmatrix(0,5,0,5);      // slow
  static double **tempm=dmatrix(0,5,0,5);  // slow
  static int n_out = 0; // limit the output
  static double non_ortho=0;

  // this is a rather complicated way of doing this though quite
  // quick -- see linear search below.
  
  // for simpler lattices this whole piece of code can be replaced
  // by a far simpler algroithm.
  
  // note that this code is implemented in 2D only here. ???
  
  static int do_once=0;
  if (!do_once){
    for (k=1; k<=3; k++) {
      m[k][1]=BravaisLattice.a[k];
      m[k][2]=BravaisLattice.b[k];
      m[k][3]=BravaisLattice.c[k];
      non_ortho+=fabs(m[k][1]*m[k][2])
                +fabs(m[k][2]*m[k][3])
                +fabs(m[k][3]*m[k][1]);
    }
    if (non_ortho)
      printf("BravaisLattice is not orthogonal, PG is slower\n");
    do_once=1;
  }

  index[4]=-1;
  for(i=1;i<=UnitCell.number;i++) {

    s[1]=x-UnitCell.x[i];
    s[2]=y-UnitCell.y[i];
    s[3]=z-UnitCell.z[i];

#if 1 // This code seems to unneccessary, but cost most of the time (Joerg)
      // only needed if Bravais is non-orthogonal (ToDo: check that)
    if (non_ortho) { // Joerg Sep05, SpeedUp for dim=1!
      // tempm is work space and will be destroyed
      for(j=1;j<=3;j++)
        for(k=1;k<=3;k++)
  	  tempm[j][k]=m[j][k];

      // what happens here? Joerg
      // input s[3],
      //       tempm=Bravais+workspace (just 1 for LC)
      ludcmp(tempm,3,ind,&d); // ToDo: most cpu, replace or make Dim-dependent
      lubksb(tempm,3,ind,s);
      // input + output is s[3]

      if (Control.output > 4)
      if (n_out<10)  // just to see if ludcmp call is really needed (1D=no)
      if (fabs(s[1]-(x-UnitCell.x[i]))
         +fabs(s[2]-(y-UnitCell.y[i]))
         +fabs(s[3]-(z-UnitCell.z[i]))>0.0) { n_out++;
        printf("findGeneralisedPositionFromCartesians i=%d n_out=%d:\n",i,n_out);
        printf("before luXX: ");
        printf(" %9.5f",x-UnitCell.x[i]);
        printf(" %9.5f",y-UnitCell.y[i]);
        printf(" %9.5f",z-UnitCell.z[i]);
        printf("\n");
        printf("after  luXX: ");
        printf(" %9.5f",s[1]);
        printf(" %9.5f",s[2]);
        printf(" %9.5f",s[3]);
        printf("\n");
      }
    }
#endif

    // check if the values for s[] are integers.

    for (k=1; k<=3; k++) {
      index[k] = ((s[k]<0)?-1:+1) * (int) (fabs(s[k])+0.5);
    }

    index[4]=i;   // UnitCell site index

    distance =fabs(x-index[1]*BravaisLattice.a[1]
                    -index[2]*BravaisLattice.b[1]
	 	    -index[3]*BravaisLattice.c[1]-UnitCell.x[i]);
    distance+=fabs(y-index[1]*BravaisLattice.a[2]
                    -index[2]*BravaisLattice.b[2]
		    -index[3]*BravaisLattice.c[2]-UnitCell.y[i]);
    distance+=fabs(z-index[1]*BravaisLattice.a[3]
                    -index[2]*BravaisLattice.b[3]
		    -index[3]*BravaisLattice.c[3]-UnitCell.z[i]);
    if (mindist < 0 || distance < mindist) {
       mindist=distance;
       configuration[1]=index[1];
       configuration[2]=index[2];
       configuration[3]=index[3];
       configuration[4]=index[4];
    }
    if (distance < 0.001) break;   // if integers, its found
  }  // i = 1...UnitCell

  if (mindist >= 0.001) {  // if no integer 
    printf("Error findGeneralisedPositionFromCartesians: no UC-site found\n"
           " out: (% 2d % 2d % 2d %d), mindist= %e\n"
           " in: xyz= % 2.0f % 2.0f % 2.0f  non_ortho= %f\n", 
           configuration[1], configuration[2], configuration[3],
           configuration[4], mindist, x, y, z, non_ortho);
    return -1;  // not found
  }

//  free_dmatrix(m,0,5,0,5);
//  free_dmatrix(tempm,0,5,0,5);
//  free_ivector(ind,0,5);
//  free_dvector(s,0,5);
  return 0;

}


// in:  configuration[5]
// out: x, y, z
// called by bookKeeping.C and define.C
void findCartesiansFromGeneralisedPosition(double &x, 
					   double &y, 
					   double &z, 
					   int *configuration)
{

  // this routine finds the cartesian position from the 
  // `generalised' lattice position => bravais vector + primitive
  // cell vector

  x=BravaisLattice.a[1]*configuration[1]+
    BravaisLattice.b[1]*configuration[2]+
    BravaisLattice.c[1]*configuration[3]+
            UnitCell.x[configuration[4]];

  y=BravaisLattice.a[2]*configuration[1]+
    BravaisLattice.b[2]*configuration[2]+
    BravaisLattice.c[2]*configuration[3]+
             UnitCell.y[configuration[4]];

  z=BravaisLattice.a[3]*configuration[1]+
    BravaisLattice.b[3]*configuration[2]+
    BravaisLattice.c[3]*configuration[3]+
             UnitCell.z[configuration[4]];

}


// coding the configuration into a descriptor index (splitted to integers)
// out: descriptor[1..MAX_SPIN_LIST]=...,999999(=invalid)
// in:  config[1..order][1..4]=0..btLength for [1..3], ??? for [4]
// called by findUniqueDescriptorOfCluster() in bookKeeping.C
// not a good name I think (Joerg05)
void findConfigurationToDescriptor(int *descriptor,
				   int (*config)[5],
				   int order)
{
  int i,j,sum_abs;
  int index;
  int tempNum;
  int block;
  static int n_out=0;

  for(i=1;i<MAX_SPIN_LIST;i++) descriptor[i]=0;  // reset descriptor bits
  
  // the first part is to determine a unique lattice index in the bounding
  // triangle for each spin flip operator in a given cluster. Each spin
  // flip is denoted by it bravais lattice vector + its number in the 
  // unit cell. 
  
  // each cell in the the bounding triangle is a unit cell.

  for(i=1;i<=order;i++) {

    // check if within bounding triangle in 1D, 2D and 3D
    for (sum_abs=0, j=1; j<=Params.dim; j++) sum_abs += abs(config[i][j]);
    if (sum_abs > Params.btLength-1  ) {
      // point lies outside the bounding triangle.
      // happens only for term27+term30 where a site is added 
      for(j=1;j<MAX_SPIN_LIST;j++) descriptor[j]=999999;  /* Sep05 */
      if (Control.output > 10) {
        printf("Error findConfigurationToDescriptor:"
              " Point outside the bounding triangle (i=%d dist=%d max=%d)\n",
              i, sum_abs, Params.btLength);
        printf("Source config:\n");
        printOutConfig(config,order);
      }
      return;
    }

    if (config[i][1]>MAX_LSUBM
     || config[i][2]>MAX_LSUBM
     || config[i][3]>MAX_LSUBM) {
      printf("Error: config > MAX_LSUBM\n");
      exit(1);
    }
    // get a unique index for every possible site (see lookUpTable)

    index = BTarray[ config[i][3] ][ config[i][2] ][ config[i][1] ];

    // at this point: 0 <= index < max_numbers_cells_in_BT

    // we now have an unique site ``index'' for 1D-3D which completely
    // specifies the position of a given spin in the bounding triangle.
    // we now wish to encode this information in a given ``block'' 
    // with a given number ``tempNum'' in ``descriptor''.

    // note that block_size gives the numbers of ``spins'' the
    // each element of the descriptor describes. Normally set to
    // 20 such spins so that we can look at large values of the 
    // quantum spin number.

    index = (index)*UnitCell.number + config[i][4] - 1;

    // at this point: 0 <= index < max_number_cells_BT*UnitCell.number
    // note that: 0 <= config[][4] < UnitCell.number

    // sites: (0,0,2,3) -> block1: { [0]=2 [1]=0 [2]=1 [3]=1 }
    //                     block2: { [4]=0 [5]=0 ...         }
    block  = index / block_size;   // 0 .. MAX_SPIN_LIST-2, int position
    index  = index % block_size;   // 0 .. block_size-1,  bit position
    
    tempNum = divisors[index+1]; // (Params.s2+1)^(i-1)=(2s+1)^(i-1)

    if (block+1 >= MAX_SPIN_LIST) { // (max_BTarray*UC)/block_size
       printf("Error: add 1 to MAX_SPIN_LIST\n"); exit(1);
    }
    descriptor[block+1] += tempNum;

    if (descriptor[block+1] < 0) {
        printf("Error descr_to_config: descr[%d]=%d < 0\n"
               " index=%d block_zize=%d pattern=%d\n",
               block+1, descriptor[block+1], index, block_size, tempNum);
    }
    // (0,0,0,1)  -> idx=0     <->    lookUpTable[idx=0..] -> (x,y,z)
    // (0,0,0,uc) -> idx=uc-1 
    if (Control.output > 6 && n_out<100) { n_out++;
      printf("[%d] (%2d %2d %2d %2d) -> index = %d -> descr[%d]+=%d\n",
        i, config[i][1],config[i][2],config[i][3],config[i][4],
        index,block+1,tempNum);
    }

  }  

}

/*
 *  input:  descriptor[MAX_SPIN_LIST]=???
 *  output: configuration[flips][x,y,z,UCell]   commented by Joerg 2005
 *  called by: obtain, define via findFundamentalSpinConfigurations, ...
 * Oct05: checkBlock removed (obsolete)
 * ToDo: use descriptor[0] to encode size of cluster !?
 */
int findClusterFromDescriptor(int *descriptor, int **configuration){
  int i, j, m, index, flips;
  int temp1, temp2, temp;
  static int n_out=0;
  int x, y, z, cell;

  x=y=z=0;   // avoid compiler warnings about using noninitialized vars
  
  if (n_out<=100) n_out++;

  flips=0;  //  the number of spin flips

  for(i=1;i<MAX_SPIN_LIST;i++) {    // MAX_SPIN_LIST*BLOCKLENGTH = SITES*UC
      temp1=descriptor[i];          // block[i] of bits for site.uc
      if (temp1==0) continue;       // no flips coded in that block
      if (temp1 < 0) {
        printf("Error descr_to_config: descr[%d]=%d < 0\n", i, temp1);
        flips=0; break; // ???
      }
      if (Control.output> 6 && n_out<100)
        printf("descr_to_config descriptor[%d]=%d\n",i,descriptor[i]);

      for(j=block_size;j>0;j--) {      // default: 10
	
        //  block1: { [0]=2 [1]=0 [2]=1 [3]=1 } -> sites: (0,0,2,3)
        //  block2: { [4]=0 [5]=0 ...         }

        // decode position
	index=(j-1)+(i-1)*block_size;   // pos*nuc+uc

	// note that: 0 <= index < block_size*MAX_SPIN_LIST

	temp=index;
	index=(index)/UnitCell.number;     // a/b  site_index  0..maxSite
	cell=temp-UnitCell.number*index;   // a%b  uc_index    0..uc-1

	// 1 <= index <= (block_size*MAX_SPIN_LIST-1)/UnitCell.number 

        // extract bitpattern
	temp2=temp1/divisors[j];             // a/b   0..2s numFlips
	temp1=temp1-temp2*divisors[j];       // a%b   rest
	
        if (!temp2) continue; // skip if no flips

        if (Control.output > 6 && n_out<100)
          printf("i = %2d j = %2d index = %2d -> (%2d %2d %2d %2d) ^ %d ...",
                  i, j, index, lookUpTable[index][1],
                               lookUpTable[index][2],
                               lookUpTable[index][3], cell+1, temp2);
        x=lookUpTable[index][1];   // see definePointsInBT() input.C
        y=lookUpTable[index][2];   // i=0..MAX_BT-1
        z=lookUpTable[index][3];
        
        for (m=1;m<=temp2;m++) {  // checkBlock: 0 ... Params.s2
          flips++;
          if (flips>MAX_LSUBM) {  // Oct05:  we mean LSUBN here
            printf("Error descr_to_config: flips > MAX_LSUBM\n");
            exit(1);
          }
          configuration[flips][1]=x;
          configuration[flips][2]=y;
          configuration[flips][3]=z;
          configuration[flips][4]=cell+1;
          if (Control.output> 4 && n_out<100)
            printf(" site = %2d (%2d %2d %2d %2d)\n",flips,x,y,z,cell+1);
        }
      }
  }

  return flips;
}

// almost the same as findClusterFromDescriptor (Joerg)
void printOutClusterFromDescriptor(int *descriptor){

  int i, j, len, max[3];
  static int **conf=imatrix(0,MAX_LSUBM+1,0,5);
  static int animal[MAX_LSUBM+1][MAX_LSUBM+1];
  max[0]=max[1]=max[2]=0; /* covering cube */

  for (i=0;i<MAX_LSUBM+1;i++)
    for (j=0;j<MAX_LSUBM+1;j++)
      animal[i][j]=0;
  len = findClusterFromDescriptor(descriptor, conf); 
  printf("(x,y,z,unitcell):");
  for (i=1; i<=len; i++) {
    printf(" (%d,%d,%d,%d)", conf[i][1], conf[i][2], conf[i][3], conf[i][4]);
    if (max[0]<conf[i][1]) max[0]=conf[i][1];
    if (max[1]<conf[i][2]) max[1]=conf[i][2];
    if (max[2]<conf[i][3]) max[2]=conf[i][3];
    animal[ conf[i][2] ][ conf[i][1] ]++;
  }
  printf("\n");
  if (max[2]<1) { /* flat animal? */
      for (i=0;i<max[1]+1;i++) {
        printf(" ");
        for (j=0;j<max[0]+1;j++)
          printf("%c",((animal[i][j])?'0'+animal[i][j]:'.'));
        printf("\n");
      }
  }
}




// in: descriptor[1..(MAX_SPIN_LIST-1)] = spinconfig of 20 sites
// out: index to spinConfigurations with same descriptor
int findDescriptorIndex(int *descriptor,
			int numberElements) {

  int i;
  int t1, t2; // upper and lower border for binary search
  int middle;

  if (numberElements < 1) return 0;

#if 0 // dont happens
  for(i=1;i<MAX_SPIN_LIST;i++)
    if (descriptor[i]!=999999) break; // check for valid number != 999999
  if (i>=MAX_SPIN_LIST) return 0;     // all numbers are 999999
#endif

  // use a binary search and return the `position' of the element (if found!)
  // if NOT found then return zero.

  t1=1-1;
  t2=numberElements+1;

  // now check that elements in the `middle' do not match `descriptor'
  // Joerg Sep 05: lot of zeros are checked, replace MAX_SPIN_LIST by
  //               a calculated maximum value

  do {
    middle=(t1+t2)/2; // if t1+t2 is odd 
    for(i=MAX_SPIN_LIST-1;i>0;i--) {
      if (spinConfigurations[middle].descriptor[i]!=descriptor[i]) {
        if (spinConfigurations[middle].descriptor[i] > descriptor[i]) 
          t2=middle;
        else
          t1=middle;
        break;
      }
    }      
    if (i==0) return middle; 
  } while (t2-t1 > 1);
		       
  // old linear search routine. Useful to keep this as a check of 
  // new binary routine  

  /*
   for(j=numberElements;j>0;j--) {
     for(i=1;i<MAX_SPIN_LIST;i++)
       if (spinConfigurations[j].descriptor[i]!=descriptor[i]) break;
     if (i==MAX_SPIN_LIST) return j;
   } 
  */
  return 0; // not found
}

/* make binary bitmap spin[]={0..1,0..1,...} from index (Joerg) */
void binary(int index, int length, int *spin)
{
  int i, y, dummy=index;
  if ((int)sizeof(int)*8 < length) { printf("binary Error OVERFLOW"); exit(1); }
  // for(i=1;i<=length;i++,index>>=1) { spin[i] = (index & 1); }
  for(i=1;i<=length;i++) {
    y= dummy/2;
    spin[i]= dummy-2*y;
    dummy=y;
  }
}

/* decode trimary code spin[]={0..2,0..2,...} from index1..3 (Joerg) */
void trimary(int index1, int index2, int index3, int length, int *spin)
{
  int i,y,dummy1, dummy2, dummy3;    
  dummy1=index1; /* spins 21..30 */
  dummy2=index2; /* spins 11..20 */
  dummy3=index3; /* spins  1..10 */
  if (length>30) { printf("trimary Error OVERFLOW"); exit(1); }

  for(i=21;i<=30 && i<=length;i++) {
    y= dummy1/3;
    spin[i]=dummy1-3*y;
    dummy1=y;
  }

  for(i=11;i<=20 && i<=length;i++) {
    y= dummy2/3;
    spin[i]=dummy2-3*y;
    dummy2=y;
  }

  for(i=1;i<=10 && i<=length;i++) {
    y= dummy3/3;
    spin[i]=dummy3-3*y;
    dummy3=y;
  }
}

/* make trimary bitmap spin[]={0..3,0..3,...} from index1..3 (Joerg) */
void quartary(int index1, int index2, int index3, int length, int *spin)
{
  
  int i,y,dummy1, dummy2, dummy3;    
  dummy1=index1;
  dummy2=index2;
  dummy3=index3;

  if (length>30) { printf("quartary Error OVERFLOW"); exit(1); }
    
  for(i=21;i<=30 && i<=length;i++) {
    y= dummy1/4;
    spin[i]=dummy1-4*y;
    dummy1=y;
  }

  for(i=11;i<=20 && i<=length;i++) {
    y= dummy2/4;
    spin[i]=dummy2-4*y;
    dummy2=y;
  }

  for(i=1;i<=10 && i<=length;i++) {
    y= dummy3/4;
    spin[i]=dummy3-4*y;
    dummy3=y;
  }
}

int findDescriptorIndex2(int *descriptor,
                        int numberElements) {

  int tempLogicFlag;
  int position;
  int i;
  int tempPosition1;
  int tempPosition2;
  int middle;
  int switchFlag=0;

  if (numberElements < 1)
    return 0;

  tempLogicFlag=1;

  // Joerg: do not understand that
  // if (descriptor[1]==999999) return 0; // not the same!

  for(i=1;i<MAX_SPIN_LIST;i++) {
    if (descriptor[i]!=999999) {
      tempLogicFlag=0;
      break;
    }
  }

  if (tempLogicFlag)
    return 0;

  position=0;

  // use a binary search and return the `position' of the element (if found!)
  // if NOT found then return zero.

  tempPosition1=1;
  tempPosition2=numberElements;

  // first check that the end elements do not match  `descriptor'

  tempLogicFlag=0;

  for(i=1;i<MAX_SPIN_LIST;i++) {
    if (latticeAnimals[tempPosition1].descriptor[i]!=descriptor[i]) {
      tempLogicFlag=1;
      break;
    }
  }
  
  if (!tempLogicFlag)
    position=tempPosition1;
  
  if (tempLogicFlag) {
    tempLogicFlag=0;
    for(i=1;i<MAX_SPIN_LIST;i++) {
      if (latticeAnimals[tempPosition2].descriptor[i]!=descriptor[i]) {
        tempLogicFlag=1;
        break;
      }
    }
    if (!tempLogicFlag)
      position=tempPosition2;
  }

  // now check that elements in the `middle' do not match `descriptor'

  if (tempLogicFlag) {
    do {

      middle=(tempPosition1+tempPosition2)/2;      
      tempLogicFlag=0;
      for(i=MAX_SPIN_LIST-1;i>0;i--) {
        if (latticeAnimals[middle].descriptor[i]!=descriptor[i]) {
          if (latticeAnimals[middle].descriptor[i] > descriptor[i])
            switchFlag=1;
          else
            switchFlag=0;
          tempLogicFlag=1;
          break;
        }
      }      

      if (!tempLogicFlag) {
        position=middle;
        tempPosition2=tempPosition1;
      } else {
        if (switchFlag)
          tempPosition2=middle;
        else
          tempPosition1=middle;
      }

    } while ((tempPosition2-tempPosition1) > 1);
  }
                       
  // old linear search routine. Useful to keep this as a check of 
  // new binary routine  

  /*
   for(j=numberElements;j>0;j--) {

    tempLogicFlag=0;
    
    for(i=1;i<MAX_SPIN_LIST;i++) {
      if (latticeAnimals[j].descriptor[i]!=descriptor[i]) {
        tempLogicFlag=1;
        break;
      }
    }

    if (!tempLogicFlag) {
      position=j;
      break;
    }
    } 
  */

  return position;

}




